From 716afdf34062f372d47981de57c2b57eba8ddcff Mon Sep 17 00:00:00 2001
From: Naohiro KORIYAMA <nkoriyama@gmail.com>
Date: Mon, 30 Apr 2012 17:35:12 +0900
Subject: [PATCH 17/17] upnp: fetch all media items correctly. (#6716)

---
 modules/services_discovery/upnp.cpp |  110 +++++++++++++++++++++++++++++++----
 modules/services_discovery/upnp.hpp |    2 +-
 2 files changed, 101 insertions(+), 11 deletions(-)

diff --git a/modules/services_discovery/upnp.cpp b/modules/services_discovery/upnp.cpp
index dc239f3..0e5fc13 100644
--- a/modules/services_discovery/upnp.cpp
+++ b/modules/services_discovery/upnp.cpp
@@ -89,8 +89,15 @@ const char* xml_getChildElementAttributeValue( IXML_Element* p_parent,
                                         const char* psz_tag_name_,
                                         const char* psz_attribute_ );
 
+const char* xml_getChildElementValue( IXML_Document* p_doc,
+                                      const char*    psz_tag_name );
+
 IXML_Document* parseBrowseResult( IXML_Document* p_doc );
 
+int parseBrowseNumberReturned( IXML_Document* p_doc );
+
+int parseBrowseTotalMatches( IXML_Document* p_doc );
+
 /*
  * Initializes UPNP instance.
  */
@@ -214,16 +221,16 @@ const char* xml_getChildElementAttributeValue( IXML_Element* p_parent,
 }
 
 /*
- * Extracts the result document from a SOAP response
+ * Returns the value of a child element, or NULL on error
  */
-IXML_Document* parseBrowseResult( IXML_Document* p_doc )
+const char* xml_getChildElementValue( IXML_Document*  p_doc,
+                                      const char*     psz_tag_name_ )
 {
-    ixmlRelaxParser( 1 );
-
     if ( !p_doc ) return 0;
+    if ( !psz_tag_name_ ) return 0;
 
     IXML_NodeList* p_result_list = ixmlDocument_getElementsByTagName( p_doc,
-                                                                   "Result" );
+                                                                   psz_tag_name_ );
 
     if ( !p_result_list ) return 0;
 
@@ -236,13 +243,69 @@ IXML_Document* parseBrowseResult( IXML_Document* p_doc )
     IXML_Node* p_text_node = ixmlNode_getFirstChild( p_result_node );
     if ( !p_text_node ) return 0;
 
-    const char* psz_result_string = ixmlNode_getNodeValue( p_text_node );
+    return ixmlNode_getNodeValue( p_text_node );
+}
+
+/*
+ * Extracts the result document from a SOAP response
+ */
+IXML_Document* parseBrowseResult( IXML_Document* p_doc )
+{
+    ixmlRelaxParser( 1 );
+
+    const char* psz_result_string = xml_getChildElementValue( p_doc, "Result" );
+    if( !psz_result_string ) return 0;
 
     IXML_Document* p_browse_doc = ixmlParseBuffer( psz_result_string );
 
     return p_browse_doc;
 }
 
+/*
+ * Get the number value of NumberReturns from a SOAP response
+ */
+int parseBrowseNumberReturned( IXML_Document* p_doc )
+{
+    ixmlRelaxParser( 1 );
+
+    const char* psz_number_returned = xml_getChildElementValue( p_doc,
+                                                                "NumberReturned" );
+    if( !psz_number_returned ) return 0;
+
+    char *psz_end;
+    long l = strtol( psz_number_returned, &psz_end, 10 );
+    if( *psz_end || l < 0 || l > INT_MAX )
+    {
+        return 0;
+    }
+    else
+    {
+        return (int)l;
+    }
+}
+
+/*
+ * Get the number value of TotalMatches from a SOAP response
+ */
+int parseBrowseTotalMatches( IXML_Document* p_doc )
+{
+    ixmlRelaxParser( 1 );
+
+    const char* psz_total_matches = xml_getChildElementValue( p_doc,
+                                                              "TotalMatches" );
+    if( !psz_total_matches ) return 0;
+
+    char *psz_end;
+    long l = strtol( psz_total_matches, &psz_end, 10 );
+    if( *psz_end || l < 0 || l > INT_MAX )
+    {
+        return 0;
+    }
+    else
+    {
+        return (int)l;
+    }
+}
 
 /*
  * Handles all UPnP events
@@ -709,7 +772,7 @@ void MediaServer::fetchContents()
 
     Container* root = new Container( 0, "0", getFriendlyName() );
 
-    _fetchContents( root );
+    _fetchContents( root, 0 );
 
     _p_contents = root;
     _p_contents->setInputItem( _p_input_item );
@@ -720,7 +783,7 @@ void MediaServer::fetchContents()
 /*
  * Fetches and parses the UPNP response
  */
-bool MediaServer::_fetchContents( Container* p_parent )
+bool MediaServer::_fetchContents( Container* p_parent, int i_starting_index )
 {
     if (!p_parent)
     {
@@ -728,9 +791,21 @@ bool MediaServer::_fetchContents( Container* p_parent )
         return false;
     }
 
+    char* psz_starting_index;
+    if( asprintf( &psz_starting_index, "%d", i_starting_index ) < 0 )
+    {
+        msg_Err( _p_sd, "asprintf error:%d", i_starting_index );
+        return false;
+    }
+
     IXML_Document* p_response = _browseAction( p_parent->getObjectID(),
                                       "BrowseDirectChildren",
-                                      "*", "0", "0", "" );
+                                      "*", /* Filter */
+                                      psz_starting_index, /* StartingIndex */
+                                      "0", /* RequestedCount */
+                                      "" /* SortCriteria */
+                                      );
+    free( psz_starting_index );
     if ( !p_response )
     {
         msg_Err( _p_sd, "No response from browse() action" );
@@ -738,6 +813,12 @@ bool MediaServer::_fetchContents( Container* p_parent )
     }
 
     IXML_Document* p_result = parseBrowseResult( p_response );
+    int i_number_returned = parseBrowseNumberReturned( p_response );
+    int i_total_matches = parseBrowseTotalMatches( p_response );
+#ifndef NDEBUG
+    msg_Dbg( _p_sd, "i_starting_index[%d]i_number_returned[%d]_total_matches[%d]\n",
+             i_starting_index, i_number_returned, i_total_matches );
+#endif
     ixmlDocument_free( p_response );
 
     if ( !p_result )
@@ -777,7 +858,7 @@ bool MediaServer::_fetchContents( Container* p_parent )
 
             Container* container = new Container( p_parent, objectID, title );
             p_parent->addContainer( container );
-            _fetchContents( container );
+            _fetchContents( container, 0 );
         }
         ixmlNodeList_free( containerNodeList );
     }
@@ -833,6 +914,15 @@ bool MediaServer::_fetchContents( Container* p_parent )
     }
 
     ixmlDocument_free( p_result );
+
+    if( i_starting_index + i_number_returned < i_total_matches )
+    {
+        if( !_fetchContents( p_parent, i_starting_index + i_number_returned ) )
+        {
+            return false;
+        }
+    }
+
     return true;
 }
 
diff --git a/modules/services_discovery/upnp.hpp b/modules/services_discovery/upnp.hpp
index 935317f..2bfa4f7 100644
--- a/modules/services_discovery/upnp.hpp
+++ b/modules/services_discovery/upnp.hpp
@@ -69,7 +69,7 @@ public:
 
 private:
 
-    bool _fetchContents( Container* p_parent );
+    bool _fetchContents( Container* p_parent, int i_starting_index );
     void _buildPlaylist( Container* p_container, input_item_node_t *p_item_node );
 
     IXML_Document* _browseAction( const char*, const char*,
-- 
1.7.9.4

